@@ -52,19 +52,6 @@ log = logging.getLogger()
 
 DEFAULT_ROOT = '/'
 
-# Name of the writable user data partition label as created by
-# ubuntu-device-flash(1).
-WRITABLE_DATA_LABEL = 'writable'
-
-# name of primary root filesystem partition label as created by
-# ubuntu-device-flash(1).
-SYSTEM_DATA_A_LABEL = 'system-a'
-
-# name of primary root filesystem partition label as created by
-# ubuntu-device-flash(1). Note that this partition will
-# only be present if this is an A/B upgrade system.
-SYSTEM_DATA_B_LABEL = 'system-b'
-
 # The tar file contains a 'system/' directory with the new files
 # to apply to the real system. It may also contain a top-level
 # file called 'removed' that lists files relative to the system
@@ -237,6 +224,13 @@ def mkfs(device, fs_type, label):
     sys.exit(1)
 
 
+def touch_file(path):
+    '''
+    Create an empty file specified by path.
+    '''
+    open(path, 'w').close()
+
+
 def get_mount_details(target):
     '''
     Returns a list comprising device, filesystem type and filesystem
@@ -536,7 +530,7 @@ class Upgrader():
         completed successfully.
         '''
         file = os.path.join(self.cache_dir, self.TIMESTAMP_FILE)
-        open(file, 'w').close()
+        touch_file(file)
 
     def get_mount_target(self):
         '''
@@ -665,6 +659,10 @@ class Upgrader():
                         .format(target_type))
             return
 
+        if self.other_considered_empty():
+            log.debug('other root partition is empty')
+            self.other_is_empty = True
+
         if self.other_is_empty and not self.other_has_been_formatted:
             # We believe "other" is empty. However, it's possible that a
             # previous attempt at upgrading failed due to a power
@@ -680,7 +678,7 @@ class Upgrader():
 
         self.remount_rootfs(writable=True)
 
-        if self.other_is_empty:
+        if self.other_is_empty and not self.full_image:
             # Copy current rootfs data to the other rootfs's blank
             # partition.
             log.debug('syncing root partitions')
@@ -708,12 +706,39 @@ class Upgrader():
 
         self.remount_rootfs(writable=False)
 
+    def other_considered_empty(self):
+        '''
+        Returns True if the other rootfs should be considered empty.
+
+        See sync_partitions() for further details.
+
+        '''
+        target = self.get_mount_target()
+        channel_ini = os.path.normpath('{}/{}'.format(target,
+                                       SYSTEM_IMAGE_CHANNEL_CONFIG))
+
+        try:
+            st = os.stat(channel_ini)
+        except OSError:
+            # ENOENT
+            return True
+
+        if st.st_size == 0:
+            # The config file is zero bytes long. This is used as a
+            # marker to denote that the initial "sync system-a to
+            # system-b" was started, but was interrupted, either in the
+            # copy phase or in the s-i unpack phase. Either way, the
+            # partition must be considered empty until the config file
+            # has a size > 0.
+            return True
+
+        return False
+
     def remount_rootfs(self, writable=False):
         target = self.get_mount_target()
 
         if writable:
-            root = subprocess.check_output(
-                ["findmnt", "-o", "SOURCE", "-n", target]).strip()
+            root, _, _ = get_mount_details(target)
 
             # ro->rw so need to fsck first.
             unmount(target)
@@ -722,14 +747,6 @@ class Upgrader():
             fsck(root)
 
             mount(root, target, "rw")
-
-            channel_ini = '{}/{}'.format(target,
-                                         SYSTEM_IMAGE_CHANNEL_CONFIG)
-            channel_ini = os.path.normpath(channel_ini)
-
-            if not os.path.exists(channel_ini):
-                log.debug('root partition on {} is empty'.format(root))
-                self.other_is_empty = True
         else:
             # rw->ro so no fsck required.
             remount(target, "ro")
@@ -886,6 +903,62 @@ class Upgrader():
 
         XXX: Assumes that the other rootfs is already mounted r/w to
         mountpoint get_mount_target().
+
+        Strategy: The partition sync is only called once (when the "other"
+        rootfs is empty), to make the empty partition identical to the
+        current rootfs partition, before unpacking the latest s-i updates
+        onto it.
+
+        Since there are 2 operations to perform for the initial update
+        (copy and unpack) and since the upgrader must be resilient to
+        power outages at any point when either of the operations are
+        running, we need a way for the upgrader to detect if a previous
+        invocation failed. This is achieved by treating the s-i
+        configuration file (SYSTEM_IMAGE_CHANNEL_CONFIG) in a special
+        way.
+
+        = Background =
+
+        A rootfs is not considered complete until the file
+        SYSTEM_IMAGE_CHANNEL_CONFIG has been unpacked onto it. However,
+        that implies we cannot copy the original
+        SYSTEM_IMAGE_CHANNEL_CONFIG from the current rootfs to "other"
+        as the copy is only the first operation and we don't want that
+        "partial update" to be considered as valid.
+
+        Ideally, we would perform the copy, excluding
+        just SYSTEM_IMAGE_CHANNEL_CONFIG. However, rsync (which has
+        excellent support for copying files with exclusions) is not
+        available.
+
+        = Approach =
+
+        The current approach is:
+
+            1) Create the directory part of SYSTEM_IMAGE_CHANNEL_CONFIG
+               on "other".
+
+            2) Touch SYSTEM_IMAGE_CHANNEL_CONFIG as an empty file on
+               "other".
+
+            3) Perform an "update copy" from current to "other". Note
+               that this will not modify the pre-existing
+               SYSTEM_IMAGE_CHANNEL_CONFIG, but will conveniently correct
+               the permissions on /etc/system-image/ to match those on
+               "current" (in the case that the original permissions used
+               by this function do not match those on "current").
+
+        The second part of the operation is to unpack the s-i update onto
+        "other" (the final part of which is to unpack
+        SYSTEM_IMAGE_CHANNEL_CONFIG (which will be >0 bytes), thus
+        making that rootfs complete and valid).
+
+        = Special treatment of SYSTEM_IMAGE_CHANNEL_CONFIG =
+
+        Using this approach makes it easy to determine if the initial
+        update failed since if either SYSTEM_IMAGE_CHANNEL_CONFIG does
+        not exist or is only zero bytes, the s-i unpack must have failed
+        meaning the rootfs must be considered "empty".
         '''
         target = self.get_mount_target()
         bindmount_rootfs_dir = tempfile.mkdtemp(prefix=script_name,
@@ -895,8 +968,13 @@ class Upgrader():
         cwd = os.getcwd()
         os.chdir(bindmount_rootfs_dir)
 
-        cmd = '/bin/cp'
-        args = [cmd, '-a', '.', target]
+        channel_ini = os.path.normpath('{}/{}'.format(target,
+                                       SYSTEM_IMAGE_CHANNEL_CONFIG))
+        si_dir = os.path.dirname(channel_ini)
+        os.makedirs(si_dir)
+        touch_file(channel_ini)
+
+        args = ['/bin/cp', '-au', '.', target]
 
         log.debug('running (from directory {}): {}'
                   .format(bindmount_rootfs_dir, args))
